"""
Author: Michel Peltriaux
Organization: Struktur- und Genehmigungsdirektion Nord, Rhineland-Palatinate, Germany
Contact: michel.peltriaux@sgdnord.rlp.de
Created on: 26.10.21

"""
from abc import abstractmethod

from django.contrib.auth.models import User, Group
from django.test import TestCase, Client
from django.urls import reverse

from compensation.models import Compensation
from intervention.models import LegalData, ResponsibilityData, Intervention
from konova.management.commands.setup_data import GROUPS_DATA
from konova.models import Geometry
from user.models import UserActionLogEntry, UserAction


class BaseTestCase(TestCase):
    """ Provides reusable functionality for specialized test cases

    """
    users = None
    groups = None
    superuser = None
    user = None

    superuser_pw = "root"
    user_pw = "root"

    @abstractmethod
    def setUp(self) -> None:
        # To be implemented in the inheriting classes
        raise NotImplementedError

    def create_users(self):
        # Create superuser and regular user
        self.superuser = User.objects.create_superuser(
            username="root",
            email="root@root.com",
            password=self.superuser_pw,
        )
        self.user = User.objects.create_user(
            username="user1",
            email="user@root.com",
            password=self.user_pw
        )
        self.users = User.objects.all()

    def create_groups(self):
        # Create groups
        for group_data in GROUPS_DATA:
            name = group_data.get("name")
            Group.objects.get_or_create(
                name=name,
            )
        self.groups = Group.objects.all()

    class Meta:
        abstract = True


class BaseViewTestCase(BaseTestCase):
    """ Wraps basic test functionality, reusable for every specialized ViewTestCase

    """
    login_url = None
    intervention = None
    compensation = None

    def setUp(self) -> None:
        self.create_users()
        self.create_groups()
        self.create_dummy_intervention()
        self.create_dummy_compensation()
        self.login_url = reverse("simple-sso-login")

    def assert_url_success(self, client: Client, urls: list):
        """ Assert for all given urls a direct 200 response

        Args:
            client (Client): The performing client
            urls (list): An iterable list of urls to be checked

        Returns:

        """
        for url in urls:
            response = client.get(url)
            self.assertEqual(response.status_code, 200, msg=f"Failed for {url}")

    def assert_url_success_redirect(self, client: Client, urls: dict):
        """ Assert for all given urls a 302 response to a certain location.

        Assert the redirect being the expected behaviour.

        Args:
            client (Client): The performing client
            urls (dict): An iterable dict of (urls, redirect_to_url) pairs to be checked

        Returns:

        """
        for url, redirect_to in urls.items():
            response = client.get(url, follow=True)
            # Expect redirects to the landing page
            self.assertEqual(response.redirect_chain[0], (redirect_to, 302), msg=f"Failed for {url}")

    def assert_url_fail(self, client: Client, urls: list):
        """ Assert for all given urls a direct 302 response

        Args:
            client (Client): The performing client
            urls (list): An iterable list of urls to be checked

        Returns:

        """
        for url in urls:
            response = client.get(url)
            self.assertEqual(response.status_code, 302, msg=f"Failed for {url}")

    def create_dummy_intervention(self):
        """ Creates an intervention which can be used for tests

        Returns:

        """
        # Create dummy data
        # Create log entry
        action = UserActionLogEntry.objects.create(
            user=self.superuser,
            action=UserAction.CREATED,
        )
        # Create legal data object (without M2M laws first)
        legal_data = LegalData.objects.create()
        # Create responsible data object
        responsibility_data = ResponsibilityData.objects.create()
        geometry = Geometry.objects.create()
        # Finally create main object, holding the other objects
        self.intervention = Intervention.objects.create(
            identifier="TEST",
            title="Test_title",
            responsible=responsibility_data,
            legal=legal_data,
            created=action,
            geometry=geometry,
            comment="Test",
        )
        self.intervention.generate_access_token(make_unique=True)

    def create_dummy_compensation(self):
        """ Creates an intervention which can be used for tests

        Returns:

        """
        if self.intervention is None:
            self.create_dummy_intervention()
        # Create dummy data
        # Create log entry
        action = UserActionLogEntry.objects.create(
            user=self.superuser,
            action=UserAction.CREATED,
        )
        geometry = Geometry.objects.create()
        # Finally create main object, holding the other objects
        self.compensation = Compensation.objects.create(
            identifier="TEST",
            title="Test_title",
            intervention=self.intervention,
            created=action,
            geometry=geometry,
            comment="Test",
        )
        self.intervention.generate_access_token(make_unique=True)


class KonovaViewTestCase(BaseViewTestCase):
    """ Holds tests for all regular views, which are not app specific

    """
    def setUp(self) -> None:
        super().setUp()

        self.home_url = reverse("home")

    def test_views_logged_in_no_groups(self):
        """ Check correct status code for all requests

        Assumption: User logged in but has no groups

        Returns:

        """
        # User logged in
        client = Client()
        client.login(username=self.superuser.username, password=self.superuser_pw)
        self.superuser.groups.set([])
        success_urls = [
            self.home_url
        ]
        self.assert_url_success(client, success_urls)

    def test_views_anonymous_user(self):
        """ Check correct status code for all requests

        Assumption: User logged in but has no groups

        Returns:

        """
        # User not logged in
        client = Client()
        urls = [
            self.home_url
        ]
        self.assert_url_fail(client, urls)


class AutocompleteTestCase(BaseViewTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.atcmplt_accs = reverse("accounts-autocomplete")
        self.atcmplt_interventions = reverse("interventions-autocomplete")
        self.atcmplt_code_comp_action = reverse("codes-compensation-action-autocomplete")
        self.atcmplt_code_comp_funding = reverse("codes-compensation-funding-autocomplete")
        self.atcmplt_code_comp_biotope = reverse("codes-biotope-autocomplete")
        self.atcmplt_code_comp_law = reverse("codes-law-autocomplete")
        self.atcmplt_code_comp_process = reverse("codes-process-type-autocomplete")
        self.atcmplt_code_comp_reg_off = reverse("codes-registration-office-autocomplete")
        self.atcmplt_code_comp_cons_off = reverse("codes-conservation-office-autocomplete")

    def _test_views_anonymous_user(self):
        # ATTENTION: As of the current state of django-autocomplete-light, there is no way to check on authenticated
        # users in a way like @loing_required or anything else. The documentation considers to check on the user's
        # authentication state during get_queryset() of the call. Therefore this test method here will stay here
        # for future clarification but won't be run due to the prefix '_'
        # User not logged in
        client = Client()
        urls = [
            self.atcmplt_accs,
            self.atcmplt_interventions,
            self.atcmplt_code_comp_action,
            self.atcmplt_code_comp_funding,
            self.atcmplt_code_comp_biotope,
            self.atcmplt_code_comp_law,
            self.atcmplt_code_comp_process,
            self.atcmplt_code_comp_reg_off,
            self.atcmplt_code_comp_cons_off,
        ]
        self.assert_url_fail(client, urls)

    def test_views_logged_in_no_groups(self):
        # User logged in
        client = Client()
        client.login(username=self.superuser.username, password=self.superuser_pw)
        self.superuser.groups.set([])
        urls = [
            self.atcmplt_accs,
            self.atcmplt_interventions,
            self.atcmplt_code_comp_action,
            self.atcmplt_code_comp_funding,
            self.atcmplt_code_comp_biotope,
            self.atcmplt_code_comp_law,
            self.atcmplt_code_comp_process,
            self.atcmplt_code_comp_reg_off,
            self.atcmplt_code_comp_cons_off,
        ]
        self.assert_url_success(client, urls)